use std::{
    any::Any,
    marker::PhantomData,
    ops::{Deref, DerefMut},
    sync::Arc,
};

use parking_lot::{RwLock, RwLockReadGuard, RwLockWriteGuard};

use crate::context::ContextHandle;

use self::internal::{
    ConstTypeId, EventHandlerBuilder, SystemBuilder, TypedEventHandlerBuilder, TypedSystemBuilder,
};

pub(crate) mod internal;

pub trait System: 'static + Send + Sync {
    const DEPENDENCIES: Dependencies;
    const EVENT_HANDLERS: EventHandlers;

    fn new(ctx: ContextHandle) -> Self;

    fn start(&mut self);
}

pub type Dependencies = &'static [Dependency];

pub struct Dependency {
    pub(crate) builder: fn() -> Box<dyn SystemBuilder>,
    pub(crate) dependencies: Dependencies,
    pub(crate) event_handlers: EventHandlers,
    pub(crate) id: ConstTypeId,
}

impl Dependency {
    pub const fn new<S: System + Send + Sync>() -> Self {
        Self {
            builder: TypedSystemBuilder::<S>::get_builder,
            dependencies: S::DEPENDENCIES,
            event_handlers: S::EVENT_HANDLERS,
            id: ConstTypeId::of::<S>(),
        }
    }
}

pub type EventHandlers = &'static [EventHandler];

pub struct EventHandler {
    // handler_any: fn(impl FnPtr),
    pub(crate) handler_any: fn(),
    pub(crate) builder: fn() -> Box<dyn EventHandlerBuilder>,
}

impl EventHandler {
    pub const fn new<S: System, T: 'static>(handler: fn(&mut S, &T)) -> Self
where {
        Self {
            handler_any: unsafe { std::mem::transmute(handler) },
            builder: TypedEventHandlerBuilder::<S, T>::get_builder,
        }
    }
}

pub struct SystemRef<S: System> {
    lock: Arc<RwLock<dyn Any>>,
    marker: PhantomData<S>,
}

impl<S: System> SystemRef<S> {
    pub(crate) fn new(lock: Arc<RwLock<dyn Any>>) -> Self {
        Self {
            lock,
            marker: PhantomData,
        }
    }

    pub fn get(&self) -> SystemReadGuard<S> {
        SystemReadGuard::new(&self.lock)
    }

    pub fn get_mut(&mut self) -> SystemWriteGuard<S> {
        SystemWriteGuard::new(&self.lock)
    }
}

pub struct SystemReadGuard<'a, S: System> {
    guard: RwLockReadGuard<'a, dyn Any>,
    marker: PhantomData<S>,
}

impl<'a, S: System> SystemReadGuard<'a, S> {
    fn new(lock: &'a Arc<RwLock<dyn Any>>) -> Self {
        Self {
            guard: lock.read(),
            marker: PhantomData,
        }
    }
}

impl<'a, S: System> Deref for SystemReadGuard<'a, S> {
    type Target = S;

    fn deref(&self) -> &Self::Target {
        self.guard
            .downcast_ref()
            .expect("SystemReadGuard contained incorrect system")
    }
}

pub struct SystemWriteGuard<'a, S: System> {
    guard: RwLockWriteGuard<'a, dyn Any>,
    marker: PhantomData<fn(S)>,
}

impl<'a, S: System> SystemWriteGuard<'a, S> {
    fn new(lock: &'a Arc<RwLock<dyn Any>>) -> Self {
        Self {
            guard: lock.write(),
            marker: PhantomData,
        }
    }
}

impl<'a, S: System> Deref for SystemWriteGuard<'a, S> {
    type Target = S;

    fn deref(&self) -> &Self::Target {
        self.guard
            .downcast_ref()
            .expect("SystemReadGuard contained incorrect system")
    }
}

impl<'a, S: System> DerefMut for SystemWriteGuard<'a, S> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        self.guard
            .downcast_mut()
            .expect("SystemWriteGuard contained incorrect system")
    }
}
